"""
===========================================================================
Example of the usage for tangent space spatial filter (TSSF) based on MOABB

Author: Jiachen XU <jiachen.xu.94@gmail.com> 

===========================================================================

This example can be successfully run on following dependencies:
Sklearn version == 0.20.3
MNE version == 0.18.2
Pyriemann version == 0.2.5
MOABB version == 0.2.1
"""

import moabb
import numpy as np
import warnings

from mne.decoding import CSP
from sklearn.svm import LinearSVC as SVM
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.pipeline import make_pipeline

from pyriemann.estimation import Covariances
from pyriemann.tangentspace import TangentSpace

from moabb.datasets import MunichMI 
from moabb.paradigms import MotorImagery
from moabb.evaluations import WithinSessionEvaluation
from moabb.pipelines.features import TSSF_pattern, TSSF

moabb.set_log_level('info')
warnings.filterwarnings("ignore")
###############################################################################
# Parameter Initialization
# ------------------------
pipelines = {}
all_decomp = ['GED', 'ED']
all_logvar = [False, True]
text_logvar = ['Cov', 'Var']

nr_ds = 5 # Choose dataset
nr_comp = 4  # Number of applied (visualized) filter components 
nr_decomp = 0  # Use GED or ED to generate filter

# L2 regularized SVM classifier using grid search to find reg. coefficients
parameters = {'C': np.logspace(-2, 2, 10)}
opt_SVM = GridSearchCV(SVC(kernel='linear'), parameters)


###############################################################################
# Initializing algorithms pipeline for analysis 
# ---------------------------------------------

pipelines['CSP (' + str(nr_comp) + ') + SVM'] = make_pipeline(
    CSP(n_components=nr_comp),
    opt_SVM)
for nr_logvar in range(2):
    ts_name = 'TSSF_' + text_logvar[nr_logvar] + ' (' + str(nr_comp) + ') + '
    # Two-steps classifcation of TSSF
    pipelines[ts_name + 'Two-steps'] = make_pipeline(
        TSSF(clf_str='SVM', func='clf', n_components=nr_comp,
             decomp=all_decomp[nr_decomp], logvar=all_logvar[nr_logvar]),
        opt_SVM)
    # One-step classifcation of TSSF
    pipelines[ts_name + 'One-step'] = make_pipeline(
        TSSF(clf_str='SVM', func='clf', n_components=nr_comp,
             decomp=all_decomp[nr_decomp], logvar=all_logvar[nr_logvar]))
pipelines['TS + w/o SF + SVM + AIRM'] = make_pipeline(
    Covariances('scm'),
    TangentSpace(metric='riemann'),
    opt_SVM)

##############################################################################
# Evaluation
# ----------
paradigm = MotorImagery(fmin=8, fmax=32, events=['left_hand', 'right_hand'])
all_datasets = paradigm.datasets
datasets = all_datasets[nr_ds]

overwrite = True  # set to True if we want to overwrite cached results
evaluation = WithinSessionEvaluation(paradigm=paradigm, datasets=datasets,
                                     suffix='DS' + str(nr_ds) + '_comp_' + str(nr_comp),
                                     overwrite=overwrite)
results = evaluation.process(pipelines)
