"""
===========================================================================
Example of the usage for tangent space spatial filter (TSSF) standalone 

Author: Jiachen XU <jiachen.xu.94@gmail.com> 

===========================================================================

This example can be successfully run on following dependencies:
Sklearn version == 0.20.3
MNE version == 0.18.2
Pyriemann version == 0.2.5
MOABB version == 0.2.1
"""

import moabb
import numpy as np
import warnings

from mne.decoding import CSP
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import cross_val_score, StratifiedKFold
from sklearn.preprocessing import LabelEncoder

from pyriemann.estimation import Covariances
from pyriemann.tangentspace import TangentSpace

from moabb.datasets import MunichMI, BNCI2014001 
from moabb.paradigms import MotorImagery
from features import TSSF

def score(clf, X, y, scoring):
    cv = StratifiedKFold(n_splits=5, shuffle=True)
    le = LabelEncoder()
    y = le.fit_transform(y)
    acc = cross_val_score(clf, X, y, cv=cv, scoring=scoring)
    return acc.mean()

def load_data(subject, session):
    MI = MotorImagery(fmin=8, fmax=32, events=['left_hand', 'right_hand'])
    dataset = MunichMI()  #  BNCI2014001() # 
    X, y, metadata = MI.get_data(dataset, [subject])
    for nr_ses, name_ses in enumerate(np.unique(metadata.session).tolist()):
        if nr_ses == session:
            ix = metadata.session == name_ses
            return X[ix], y[ix]


if __name__ == "__main__":
    ###############################################################################
    # Parameter Initialization
    # ------------------------

    warnings.filterwarnings("ignore")
    pipelines = {}
    all_decomp = ['GED', 'ED']
    all_logvar = [False, True]
    text_logvar = ['Cov', 'Var']

    nr_comp = 4  # Number of applied (visualized) filter components 
    nr_decomp = 0  # Use GED or ED to generate filter

    # L2 regularized SVM classifier using grid search to find reg. coefficients
    parameters = {'C': np.logspace(-2, 2, 10)}
    opt_SVM = GridSearchCV(SVC(kernel='linear'), parameters)

    ###############################################################################
    # Initializing algorithms pipeline for analysis 
    # ---------------------------------------------

    pipelines['CSP (' + str(nr_comp) + ') + SVM'] = make_pipeline(
        CSP(n_components=nr_comp),
        opt_SVM)
    for nr_logvar in range(2):
        ts_name = 'TSSF_' + text_logvar[nr_logvar] + ' (' + str(nr_comp) + ') + '
        # Two-steps classifcation of TSSF
        pipelines[ts_name + 'Two-steps'] = make_pipeline(
            TSSF(clf_str='SVM', func='clf', n_components=nr_comp,
                 decomp=all_decomp[nr_decomp], logvar=all_logvar[nr_logvar]),
            opt_SVM)
        # One-step classifcation of TSSF
        pipelines[ts_name + 'One-step'] = make_pipeline(
            TSSF(clf_str='SVM', func='clf', n_components=nr_comp,
                 decomp=all_decomp[nr_decomp], logvar=all_logvar[nr_logvar]))
    pipelines['TS + w/o SF + SVM + AIRM'] = make_pipeline(
        Covariances('scm'),
        TangentSpace(metric='riemann'),
        opt_SVM)

    # You can replace load_data() function with your own data instead of using MOABB dependencies
    # X: float& ndarray, shape (n_trials, n_channels, n_samples)
    # y: float& ndarray, shape (n_trials,)
    X, y = load_data(subject=1, session=0)
    for name, clf in pipelines.items():
        acc = score(clf, X, y, scoring='roc_auc')
        print(name + ': '+ str(acc))

